import { declare } from '@babel/helper-plugin-utils';
import syntaxPipelineOperator from '@babel/plugin-syntax-pipeline-operator';
import { types } from '@babel/core';

function isConciseArrowExpression(node) {
  return types.isArrowFunctionExpression(node) && types.isExpression(node.body) && !node.async;
}
const buildOptimizedSequenceExpression = ({
  call,
  path,
  placeholder
}) => {
  const {
    callee: calledExpression
  } = call;
  const pipelineLeft = path.node.left;
  const assign = types.assignmentExpression("=", types.cloneNode(placeholder), pipelineLeft);
  const expressionIsArrow = isConciseArrowExpression(calledExpression);
  if (expressionIsArrow) {
    let param;
    let optimizeArrow = true;
    const {
      params
    } = calledExpression;
    if (params.length === 1 && types.isIdentifier(params[0])) {
      param = params[0];
    } else if (params.length > 0) {
      optimizeArrow = false;
    }
    if (optimizeArrow && !param) {
      return types.sequenceExpression([pipelineLeft, calledExpression.body]);
    } else if (param) {
      path.scope.push({
        id: types.cloneNode(placeholder)
      });
      path.get("right").scope.rename(param.name, placeholder.name);
      return types.sequenceExpression([assign, calledExpression.body]);
    }
  } else if (types.isIdentifier(calledExpression, {
    name: "eval"
  })) {
    const evalSequence = types.sequenceExpression([types.numericLiteral(0), calledExpression]);
    call.callee = evalSequence;
  }
  path.scope.push({
    id: types.cloneNode(placeholder)
  });
  return types.sequenceExpression([assign, call]);
};

const minimalVisitor = {
  BinaryExpression(path) {
    const {
      scope,
      node
    } = path;
    const {
      operator,
      left,
      right
    } = node;
    if (operator !== "|>") return;
    const placeholder = scope.generateUidIdentifierBasedOnNode(left);
    const call = types.callExpression(right, [types.cloneNode(placeholder)]);
    path.replaceWith(buildOptimizedSequenceExpression({
      placeholder,
      call,
      path: path
    }));
  }
};

const topicReferenceVisitor = {
  exit(path, state) {
    if (path.isTopicReference()) {
      state.topicReferences.push(path);
    } else {
      if (state.topicReferences.length === 0 && !state.sideEffectsBeforeFirstTopicReference && !path.isPure()) {
        state.sideEffectsBeforeFirstTopicReference = true;
      }
    }
  },
  "ClassBody|Function"(_, state) {
    if (state.topicReferences.length === 0) {
      state.sideEffectsBeforeFirstTopicReference = true;
    }
  }
};
const visitor = {
  BinaryExpression: {
    exit(path) {
      const {
        scope,
        node
      } = path;
      if (node.operator !== "|>") {
        return;
      }
      const pipeBodyPath = path.get("right");
      if (pipeBodyPath.node.type === "TopicReference") {
        path.replaceWith(node.left);
        return;
      }
      const visitorState = {
        topicReferences: [],
        sideEffectsBeforeFirstTopicReference: pipeBodyPath.isFunction()
      };
      pipeBodyPath.traverse(topicReferenceVisitor, visitorState);
      if (visitorState.topicReferences.length === 1 && (!visitorState.sideEffectsBeforeFirstTopicReference || path.scope.isPure(node.left, true))) {
        visitorState.topicReferences[0].replaceWith(node.left);
        path.replaceWith(node.right);
        return;
      }
      const topicVariable = scope.generateUidIdentifierBasedOnNode(node);
      scope.push({
        id: topicVariable
      });
      visitorState.topicReferences.forEach(path => path.replaceWith(types.cloneNode(topicVariable)));
      path.replaceWith(types.sequenceExpression([types.assignmentExpression("=", types.cloneNode(topicVariable), node.left), node.right]));
    }
  }
};

const fsharpVisitor = {
  BinaryExpression(path) {
    const {
      scope,
      node
    } = path;
    const {
      operator,
      left,
      right
    } = node;
    if (operator !== "|>") return;
    const placeholder = scope.generateUidIdentifierBasedOnNode(left);
    const call = right.type === "AwaitExpression" ? types.awaitExpression(types.cloneNode(placeholder)) : types.callExpression(right, [types.cloneNode(placeholder)]);
    const sequence = buildOptimizedSequenceExpression({
      placeholder,
      call,
      path: path
    });
    path.replaceWith(sequence);
  }
};

const updateTopicReferenceVisitor = {
  PipelinePrimaryTopicReference(path) {
    path.replaceWith(types.cloneNode(this.topicId));
  },
  PipelineTopicExpression(path) {
    path.skip();
  }
};
const smartVisitor = {
  BinaryExpression(path) {
    const {
      scope
    } = path;
    const {
      node
    } = path;
    const {
      operator,
      left,
      right
    } = node;
    if (operator !== "|>") return;
    const placeholder = scope.generateUidIdentifierBasedOnNode(left);
    scope.push({
      id: placeholder
    });
    let call;
    if (types.isPipelineTopicExpression(right)) {
      path.get("right").traverse(updateTopicReferenceVisitor, {
        topicId: placeholder
      });
      call = right.expression;
    } else {
      let callee = right.callee;
      if (types.isIdentifier(callee, {
        name: "eval"
      })) {
        callee = types.sequenceExpression([types.numericLiteral(0), callee]);
      }
      call = types.callExpression(callee, [types.cloneNode(placeholder)]);
    }
    path.replaceWith(types.sequenceExpression([types.assignmentExpression("=", types.cloneNode(placeholder), left), call]));
  }
};

const visitorsPerProposal = {
  minimal: minimalVisitor,
  hack: visitor,
  fsharp: fsharpVisitor,
  smart: smartVisitor
};
var index = declare((api, options) => {
  api.assertVersion("8.0.0-alpha.12");
  const {
    proposal
  } = options;
  if (proposal === "smart") {
    console.warn(`The smart-mix pipe operator is deprecated. Use "proposal": "hack" instead.`);
  }
  return {
    name: "proposal-pipeline-operator",
    inherits: syntaxPipelineOperator,
    visitor: visitorsPerProposal[options.proposal]
  };
});

export { index as default };
//# sourceMappingURL=index.js.map
